#if !defined(__COMP831_WIN32_COMM_H__)
#define __COMP831_WIN32_COMM_H__

#define USECOMM			// yes, we need the COMM API

#include <windows.h>
#include <string.h>
#include <io.h>
#include <memory.h>

// hard coded maximum number of ports for device under Win32

#define MAXPORTS			4

#define RXQUEUE				4096
#define TXQUEUE				4096

// Flow control flags

#define FC_DTRDSR			0x01
#define FC_RTSCTS			0x02
#define FC_XONXOFF			0x04

// ascii definitions

#define ASCII_BEL			0x07
#define ASCII_BS			0x08
#define ASCII_LF			0x0A
#define ASCII_CR			0x0D
#define ASCII_XON			0x11
#define ASCII_XOFF			0x13

// data structures

typedef struct tagCOMMINFO
{
	HANDLE			idComDev;
	BYTE			bPort;
	BOOL			fConnected;
	BOOL			fXonXoff;
	BOOL			fUseCNReceive;
	BOOL			fDisplayErrors;
	BYTE			bByteSize;
	BYTE			bFlowCtrl;
	BYTE			bParity;
	BYTE			bStopBits;
	DWORD	 		dwBaudRate;
	HANDLE			hPostEvent;
	HANDLE			hWatchThread;
	HANDLE			hWatchEvent;
	DWORD			dwThreadID;
	OVERLAPPED		osWrite;
	OVERLAPPED		osRead;
	HWND			hWnd;

} COMMINFO, * LPCOMMINFO;

DWORD	BaudTable[] =
{
	CBR_110, CBR_300, CBR_600, CBR_1200, CBR_2400,
	CBR_4800, CBR_9600, CBR_14400, CBR_19200, CBR_38400,
	CBR_56000, CBR_128000, CBR_256000
};

DWORD	ParityTable[] =
{
	NOPARITY, EVENPARITY, ODDPARITY, MARKPARITY, SPACEPARITY
};

DWORD	StopBitsTable[] =
{
	ONESTOPBIT, ONE5STOPBITS, TWOSTOPBITS
};

// message delivered to window when comm data is received

#define MAXBLOCK	80

struct tagCOMMDATA
{
	int nLength ;
	BYTE abIn[ MAXBLOCK + 1] ;
};

typedef struct tagCOMMDATA COMMDATA, * LPCOMMDATA;

#define MNO_COMMDATA			(WM_USER + 1000)

/*
 *	wParam = 0
 *	lParam = LPCOMMDATA
 *		pointer to COMMDATA structure containing the
 *		received data, dynamically allocated using LocalAlloc();
 *		user may free this struct when finished using LocalFree()
 */

// Public Functions in comm.c

LPCOMMINFO InitCommInfo( HWND hWnd );
VOID DestroyCommInfo( LPCOMMINFO lpCom );
BOOL OpenConnection( LPCOMMINFO lpCom );
BOOL SetupConnection( LPCOMMINFO lpCom );
BOOL CloseConnection( LPCOMMINFO lpCom );
INT ReadCommBlock( LPCOMMINFO lpCom, LPSTR lpszBlock, INT nMaxLength );
BOOL WriteCommBlock( LPCOMMINFO lpCom, LPSTR lpByte, DWORD dwBytesToWrite );

// Private Functions in comm.c

VOID Fatal( LPCTSTR szErrorMessage, LPCTSTR szCaption );
VOID DisplayCommError( DWORD dwError, LPCOMMINFO lpCom );
INT HandleReadCommError( LPCOMMINFO lpCom, LPCOMSTAT lpComStat );
BOOL HandleWriteCommError( LPCOMMINFO lpCom, LPCOMSTAT lpComStat );
VOID CommWatchNotify( HWND hWnd, LPCOMMDATA lpData );
VOID CommWatchRead( LPCOMMINFO lpCom );
DWORD WINAPI CommWatchProc( LPVOID lpData );

#endif /* __COMP831_WIN32_COMM_H__ */
